// upc_link.h
// This header provides utilities for managing interoperability with the Berkeley UPC Runtime (aka upcr)

#ifndef __UPC_LINK_H

#if !NOBS_DISCOVERY
#include <gasnet.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif

// returns non-zero iff Berkeley UPC Runtime was linked in
// otherwise, returns 0 and all other functions declared in this header have undefined behavior.
extern int upcxx_upc_is_linked(void);

// Initialize the Berkeley UPC runtime system (if needed) 
// and return the GASNet-EX objects which are SHARED with upcr
extern void upcxx_upc_init(
                gex_Client_t           *client_p,
                gex_EP_t               *ep_p,
                gex_TM_t               *tm_p
            );

// allocate sz bytes from the libupcr shared heap
// the _all variant is collective and sz must be single-valued
// out-of-memory failures are fatal.
extern void *upcxx_upc_alloc(size_t sz);
extern void *upcxx_upc_all_alloc(size_t sz);

// release storage previously allocated using upcxx_upc_alloc()
extern void upcxx_upc_free(void *ptr);
// collectively release storage previously allocated using upcxx_upc_all_alloc()
extern void upcxx_upc_all_free(void *ptr);

#ifdef __cplusplus
} // extern "C"
#endif

#endif // __UPC_LINK_H
